/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares types for maps from supported interface identifier
 * to the identifier for the extension interface through which
 * we will provides that support.
 *
 * For example, we might support two historical versions of the
 * "vrpsx_system" interface in addition to the one we directly
 * implement.  If each one is backward compatible with the ones
 * that came before it--in the binary sense of compatibility--then
 * we would map from the identifiers for each of those three
 * interfaces to identifier for the newest interface.
 */

#ifndef VRPSX_DIRECT_SUPPORT_MAP_H
#define VRPSX_DIRECT_SUPPORT_MAP_H 1

#include <vrpsx_embedding.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* An array of with this element type provides the support map
 * for an object supporting vrpsx_identified_t.
 *
 * The "supported" member is the lookup key; the array
 * must be sorted and unique with respect to this member.
 */
typedef struct object_binding_struct {
    /* The identifier for an interface supported by the object in question.
     */
    vrpsx_interface_identifier_t supported;

    /* The identifier for the interface that will be provided when
     * (supported) is requested; this interface must extend (supported).
     */
    vrpsx_interface_identifier_t provided;

    /* The offset in bytes of (provided) from the start of the object.
     */
    size_t offset;
} object_binding_t;


/* An array of with this element type provides the support map
 * for a program that can call functions of identified type.
 *
 * The "supported" member is the lookup key; the array
 * must be sorted and unique with respect to this member.
 */
typedef struct function_binding_struct {
    /* The identifier for a supported function interface;
     * it is supported in the sense that the program can
     * call functions expecting that interface and having
     * the function type specified by that interface.
     */
    vrpsx_interface_identifier_t supported;

    /* The identifier for the function interface whose type will actually
     * be used to call functions of the type specified by (supported).
     *
     * The function interface (provided) must extend (supported) in
     * the sense that any function expecting the (supported) interface
     * and having the function type that it specifies must be safely
     * callable as if it expected (provided) and had the function type
     * specified by (provided).
     */
    vrpsx_interface_identifier_t provided;
} function_binding_t;


extern vrpsx_interface_identifier_t const *object_supports(
    object_binding_t const support_map[],
    size_t dimension,
    vrpsx_interface_identifier_t interface_identifier,
    size_t *offset);

extern vrpsx_interface_identifier_t const *caller_supports(
    function_binding_t const support_map[],
    size_t dimension,
    vrpsx_interface_identifier_t interface_identifier);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_DIRECT_SUPPORT_MAP_H */
