/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * When combined with a calling convention, the following declarations
 * define a binary interface for binding identifiers to interfaces and
 * returning error information.
 *
 * Interface Identifier:
 *
 * The interface defined by this header, and all extensions thereof,
 * provide their identifiers through this interface, and may also
 * provide other supported interfaces, indexed by identifier.
 *
 * This interface uses 32-bit unsigned integers as interface identifiers.
 * When presenting such an interface identifier in a textual context we
 * recommend its lowercase hexadecimal representation, with all 8 digits
 * even if there are leading 0s, and without any radix prefix, except in
 * contexts that demand it, such as a C integer literal.
 *
 * We need the identifier for an interface to differ from the identifiers for
 * inequivalent interfaces, even when one is backward compatible with the
 * other.  For more on interface compatibility, see "Compatibility" (below).
 *
 * In order to avoid needing a centralized authority for interface identifiers,
 * we recommend that you assign identifiers in a random or pseudorandom fashion
 * that minimizes the chance of a collision (see "Chance of Collision").
 *
 * When declaring compatibility with another interface, it is assumed that you
 * have some knowledge of the identifiers that other implementations used with
 * that interface.  You can use some or all of those identifiers, even if they
 * were chosen in a different way than you choose your identifier.
 *
 * Pseudorandom Identification:
 *
 * We suggest the following pseudorandom approach, which has the advantage of
 * automatically detecting changes in your interface, though of course it has
 * no automatic way to detect which changes are compatible.  Check for a tool
 * accompanying this header that may implement this approach.
 *
 * Define the "recursive digest" of a header to be the MD5 digest of
 * the UTF-8 text of the header after the following transformations
 * have been applied, in order:
 *
 *   1. Replace every maximal contiguous sequence of one or more
 *      ' ', '\t', '\r', or '\n' characters with "\n" if it
 *      contains '\r' or '\n' and by " " if it contains only
 *      ' ' and '\t', except inside of C string/character
 *      literals and include pathnames.
 *
 *   2. Split the text into lines at '\n' characters that are outside
 *      of C string/character literals.  (Every '\n' in a literal is
 *      treated as an ordinary character when we talk about "lines".)
 *
 *   3. Delete any ' ' character at the beginning of every line, and at the
 *      end of every line.  (It cannot be inside a string/character literal.)
 *
 *   4. Delete any initial and terminal contiguous sequence of empty lines.
 *
 *   5. Delete any ' ' following a '#' at the start of a line,
 *      except inside of C string/character literals and comments.
 *
 *   6. For any include directive whose line does not contain
 *      "EXCLUDE_FROM_INTERFACE_IDENTIFIER", replace the pathname and its
 *      delimiters (<...> or "...") with the recursive digest of that header,
 *      expressed in 32-digit big-endian lowercase hexadecimal form.
 *      If that recursive digest is undefined thanks to an inclusion cycle,
 *      then an error results--we require that any cyclic inclusions be
 *      excluded using "EXCLUDE_FROM_INTERFACE_IDENTIFIER" or avoided by
 *      merging headers for interdependent interfaces, as appropriate.
 *
 *   7. Append '\n' to every line.
 *
 *   8. Concatenate the lines in order.
 *
 * Each transformation assumes that you have already performed the
 * previous transformation.  However, the above steps merely define
 * "recursive digest" mathematically; you may compute recursive digests
 * using any implementation of any functionally equivalent algorithm.
 *
 * Now suppose that an interface named N is defined primarily in header H.
 * Ordinarily N should be the UTF-8 base name for the C constructs that
 * define the interface: for example, "vrpsx_identified" and "vrpsx_error"
 * name the two identified interfaces defined by this header.  Compute the
 * recursive digest of H, as defined above, express it in 32-digit big-endian
 * lowercase hexadecimal form, append ' ' followed by N, and compute the
 * MD5 digest of that space-separated pair.  Take the last 4 octets of
 * that digest as the 32-bit big-endian unsigned integer representation
 * of the interface identifier for N.  (We choose a trailing contiguous
 * subsequence of octets with the idea that a wider interface identifier
 * could be defined by just increasing the length of that subsequence,
 * and the resulting identifiers would be partially compatible in the
 * sense that their low bits would agree with the 32-bit formula.)
 *
 * Typically it is best to define at most one identified interface per header,
 * in order to avoid unnecessary changes to identifiers when irrelevant changes
 * are made.  However, true cyclic dependencies between identified interfaces
 * make that impractical.  For example, "vrpsx_identified" and "vrpsx_error"
 * are interdependent and therefore should be defined in the same header.
 *
 * Chance of Collision:
 *
 * Recall that one option is to assign identifiers at random.  As long
 * as you find some way to remember to make new random assignments when
 * you create new variants of interfaces, that is perfectly valid.
 *
 * When picking identifiers at random, the chance of assigning the same
 * interface identifer to two inequivalent interfaces can be understood
 * the answer to a "Birthday Problem".  See
 *
 *     http://en.wikipedia.org/wiki/Birthday_problem
 *
 * for a full treatment, but given a pool of 100 interfaces, the chance that
 * two interfaces are assigned the same identifier should be about 1e-6, and
 * a pool of 10000 interfaces has about a one percent collision chance.
 *
 * Though we are very far from a proof, it seems a reasonable guess that the
 * pseudorandom approach suggested above would have similar collision chances.
 *
 * If you are using the fully random approach to assigning identifiers, then
 * please retry your random choice if you notice a collision with an existing
 * identifier for a inequivalent interface.  (For a *completely* equivalent
 * interface you can just reuse the existing identifier.)
 *
 * If you are using the pseudorandom approach and notice a collision with an
 * existing identifier for a inequivalent interface, when we recommend making
 * some trivial change to the header that you are introducing (such as adding
 * a comment that mentions the collision).
 *
 * Unfortunately, if preexisting implementations assigned the same identifier
 * to inequivalent interfaces, then you cannot be simultaneously compatible
 * with both, though of course you could have a mode switch.  For this reason
 * we encourage implementors to introduce new identifiers whenever there is
 * any doubt at all about equivalence.  The cost of a new identifier is small
 * because of compatibility features (see "Compatibility").
 *
 * But if the number of interfaces spins out of control then one could consider
 * defining an analogous interface with a much longer interface identifier.
 * One could use this interface to obtain that wider interface when available.
 *
 * Compatibility:
 *
 * Note that headers defining equivalent interfaces may nevertheless have
 * differing identifiers.  Therefore when a client of this interface uses it to
 * request an interface, we explicitly allow the implementation to provide any
 * interface that is backward compatible with the requested interface, even if
 * it has a different identifier.  Of course equivalent interfaces are backward
 * compatible with each other, by definition.
 *
 * A client should respect the choice of an implementation to provide an
 * interface with a different identifier than the one requested.  Unless
 * working around a defect in an old implementation, a client should trust
 * that the provided interface really is compatible, just as it would by
 * default trust any act taken by an implementation on its behalf.
 *
 * In particular, when a client requests an interface and receives one
 * whose identifier is not known to the client, then it is likely that the
 * implementation is newer than the client, and if any workarounds are
 * needed then they are probably in the implementation, not the client.
 *
 * On the other hand, it is perfectly reasonable for a client to request
 * its most preferred interface first, and if that is not available then
 * then request a more broadly compatible alternative.
 *
 * Backward compatibility is at the binary call level, NOT the source level.
 * Furthermore, an interface may be backward compatible with another on some
 * platforms but not others, due to integer size issues, etc.
 *
 * Note that comments DO affect the generated identifier because they help to
 * define binary interface compatibility.  For example, if you change a comment
 * to say that a NULL argument is allowed where previously NULL was disallowed,
 * or perhaps not even mentioned clearly, then you are extending the interface
 * to support additional client behavior and therefore it should have a new
 * identifier, even though it is backward compatible.
 */

#ifndef VRPSX_IDENTIFIED_H
#define VRPSX_IDENTIFIED_H 1

#include "vrpsx_reference_counted.h"
#include "vrpsx_stdint.h" /* EXCLUDE_FROM_INTERFACE_IDENTIFIER because binary
                           * compatibility is affected only by the actual size
                           * of binary integer we choose, not which C type
                           * expresses a particular size of binary integer.
                           */
#include <stddef.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER because basic
                     * definitions such as size_t are part of the architecture,
                     * and we are not attempting binary compatibility between
                     * different architectures (such as calling from 32-bit
                     * code into 64-bit code).
                     */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* The binary representation of an interface identifer.  Must agree with the
 * "vrpsx_identify" program, though ideally we will never change this type.
 */
typedef vrpsx_uint32_t vrpsx_interface_identifier_t;


typedef struct vrpsx_error_struct vrpsx_error_t;


/* An interface through which you can obtain other interfaces by identifier.
 *
 * Interface Lifetime: implied by the context, or by an interface extension.
 * Thread Safety: thread-safe.
 */
typedef struct vrpsx_identified_struct {
#define VRPSX_IDENTIFIED_MEMBERS(VTBL)          \
    VTBL const *vtbl;

    VRPSX_IDENTIFIED_MEMBERS(struct vrpsx_identified_vtbl_struct)
} vrpsx_identified_t;

typedef struct vrpsx_identified_end_struct {
} vrpsx_identified_end_t;

typedef struct vrpsx_identified_vtbl_struct {
#define VRPSX_IDENTIFIED_VTBL_MEMBERS(TYPE)                             \
                                                                        \
    /* The identifier for this interface--usually an                    \
     * extension of the one defined in this header.                     \
     */                                                                 \
    vrpsx_interface_identifier_t interface_identifier;                  \
                                                                        \
    /* NOTE: Unlike most interface operations, the first argument is    \
     * not a pointer to this interface; rather, it is a pointer to      \
     * such a pointer.  Both the argument and the interface pointer     \
     * must be non-NULL when you call this operation.                   \
     *                                                                  \
     * If an alternative interface can be made available that is        \
     * backward compatible with the specified interface (even if it is  \
     * incompatible with this interface), then this operation stores    \
     * its address in (*out_alternative), stores NULL in (*identified), \
     * and stores NULL in (*out_error).                                 \
     *                                                                  \
     * WARNING: The caller must own a reference to this interface in    \
     * order to call this operation, because on success the caller      \
     * loses ownership of the old reference and acquires ownership      \
     * of the new reference.  Thereafter the caller must follow the     \
     * lifetime control policy specified by the new interface.  If      \
     * the old interface was reference counted or otherwise multiply    \
     * owned, then the other owners are unaffected--they may continue   \
     * to use their references to the old interface.  In this success   \
     * case *out_error is set to NULL (its old value is ignored).       \
     *                                                                  \
     * The identifier of the resulting interface may differ from the    \
     * requested one as long as backward compatibility is expected.     \
     * Here we speak of binary compatibility, not source compatibility. \
     *                                                                  \
     * But if no compatible interface is supported, or if there is      \
     * some other error in making the interface available, then this    \
     * operation stores NULL in *out_alternative, this operation        \
     * leaves (*identified) unchanged, and the caller retains           \
     * ownership of that reference to this interface.  In this case     \
     * this operation stores an appropriate error interface in          \
     * *out_error, transferring to the caller ownership of one          \
     * reference to that error interface.                               \
     *                                                                  \
     * The most portable usage is to pass the address of a variable of  \
     * type (void *) as the value of "out_alternative", then cast the   \
     * resulting value of that variable to the appropriate interface    \
     * type.  (On some platforms such casts change the binary value,    \
     * though that is rare at the moment.)                              \
     *                                                                  \
     * The old values of *out_alternative and *out_error are ignored.   \
     */                                                                 \
    void (*replace_interface)(                                          \
        TYPE **identified,                                              \
        vrpsx_interface_identifier_t interface_identifier,              \
        void **out_alternative,                                         \
        vrpsx_error_t **out_error);                                     \
                                                                        \
    /* Marks the end of this interface, after which come any extension  \
     * operations.  This is not just for documentation: this member     \
     * may trigger a compiler warning or error if you make a mistake    \
     * while listing methods, though you may need to enable pointer     \
     * incompatibility checks in your compiler.                         \
     */                                                                 \
    vrpsx_identified_end_t const *identified_end;

    VRPSX_IDENTIFIED_VTBL_MEMBERS(vrpsx_identified_t)
} vrpsx_identified_vtbl_t;


/* An identified interface whose lifetime is controlled by reference counting.
 *
 * You can cast a pointer to this interface to (vrpsx_identified_t *)
 * because this interface extends vrpsx_identified_t.
 *
 * Extends: vrpsx_identified_t.
 * Interface Lifetime: reference counted.
 * Thread Safety: thread-safe.
 * See Also: vrpsx_reference_counted_t.
 */
typedef struct vrpsx_counted_identified_struct {
#define VRPSX_COUNTED_IDENTIFIED_MEMBERS(VTBL)  \
    VRPSX_IDENTIFIED_MEMBERS(VTBL)

    VRPSX_COUNTED_IDENTIFIED_MEMBERS(
        struct vrpsx_counted_identified_vtbl_struct)
} vrpsx_counted_identified_t;

typedef struct vrpsx_counted_identified_end_struct {
} vrpsx_counted_identified_end_t;

typedef struct vrpsx_counted_identified_vtbl_struct {
#define VRPSX_COUNTED_IDENTIFIED_VTBL_MEMBERS(TYPE)                     \
    VRPSX_IDENTIFIED_VTBL_MEMBERS(TYPE)                                 \
                                                                        \
    /* NOTE: Unlike most interface operations, the first argument is    \
     * not a pointer to this interface; rather, it is a pointer to      \
     * such a pointer.  Both the argument and the interface pointer     \
     * must be non-NULL when you call this operation.                   \
     *                                                                  \
     * This operation sets (*counted) to NULL, atomically decrements    \
     * the reference count for this interface, and if it drops to 0     \
     * then releases any resources associated with this interface.      \
     *                                                                  \
     * If implementation resources are associated with more than one    \
     * interface, then the implementation is responsible for ensuring   \
     * their appropriately timed and reasonably prompt cleanup.         \
     *                                                                  \
     * WARNING: Implementations are free to maintain per-interface      \
     * reference counts if desired, and dispose of interfaces based     \
     * upon those counts.  Therefore an interface client must release   \
     * its reference to an interface THROUGH THAT SAME INTERFACE        \
     * (unless some interface extension clearly documents that the      \
     * implementation maintains just a one reference count for          \
     * all of the interfaces that it supports).                         \
     */                                                                 \
    void (*release)(                                                    \
        TYPE **counted);                                                \
                                                                        \
    /* Atomically increments the reference count for this interface.    \
     */                                                                 \
    void (*acquire)(                                                    \
        TYPE *counted);                                                 \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    vrpsx_counted_identified_end_t const *counted_identified_end;

    VRPSX_COUNTED_IDENTIFIED_VTBL_MEMBERS(vrpsx_counted_identified_t);
} vrpsx_counted_identified_vtbl_t;


/* An interface to an error.
 *
 * You can cast a pointer to this interface to (vrpsx_counted_identified_t *)
 * because this interface extends vrpsx_counted_identified_t.
 *
 * An operation that uses an error interface pointer to describe any failures
 * it might have must provide a NULL pointer when and only when it succeeds.
 * On failure, ownership of one error interface reference passes to the caller.
 *
 * (Operations whose purpose is to manipulate or list information concerning
 * errors other than those that occur in the operation itself may provide
 * error interface pointers that do NOT describe their own failures, and
 * those interfaces are not subject to this general policy.)
 *
 * Implementations that produce error interfaces must degrade gracefully under
 * low memory conditions, when internationalization files are not present, etc.
 * They must return some non-NULL error interface pointer, though as necessary
 * they may substitute untranslated or statically allocated alternative text.
 *
 * Extends: vrpsx_counted_identified_t.
 * Interface Lifetime: reference counted.
 * Thread Safety: thread-safe.
 */
struct vrpsx_error_struct {
#define VRPSX_ERROR_MEMBERS(VTBL)               \
    VRPSX_COUNTED_IDENTIFIED_MEMBERS(VTBL)

    VRPSX_ERROR_MEMBERS(struct vrpsx_error_vtbl_struct)
};

typedef struct vrpsx_error_end_struct {
} vrpsx_error_end_t;

typedef struct vrpsx_error_vtbl_struct {
#define VRPSX_ERROR_VTBL_MEMBERS(TYPE)                                  \
    VRPSX_COUNTED_IDENTIFIED_VTBL_MEMBERS(TYPE)                         \
                                                                        \
    /* Returns a non-NULL NUL-terminated human-readable error message.  \
     *                                                                  \
     * The usable lifetime of the returned string is bounded by both    \
     * the lifetime of this interface and the next operation of this    \
     * or any alternative interface that could modify that text for     \
     * some argument values (not necessarily the ones actually passed). \
     */                                                                 \
    char const *(*text)(                                                \
        TYPE const *error);                                             \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    vrpsx_error_end_t const *error_end;

    VRPSX_ERROR_VTBL_MEMBERS(vrpsx_error_t);
} vrpsx_error_vtbl_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_IDENTIFIED_H */
