/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * An buffered interface to an output stream.
 */

#ifndef VRPSX_OSTREAMBUF_H
#define VRPSX_OSTREAMBUF_H 1

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "vrpsx_identified.h"


/* An output stream buffer, similar to the C++ std::streambuf.
 *
 * Conceptual Stream:
 *
 * Conceptually, this interface incrementally receives a sequence of
 * char values and takes responsibility for eventually delivering the
 * entire sequence to an implementation-defined destination.  These
 * char values may express text or binary data.
 *
 * The client can append to the conceptual stream both through the buffer (see
 * "Buffer Management") and through the data arguments of certain operations.
 *
 * Errors:
 *
 * An implementation might fail to deliver some or all of the sequence to its
 * destination.  The "error" member is 0 until the first such failure, after
 * which it holds the error number for that first failure.  (Other kinds of
 * errors, such as invalid arguments, may be transient.)
 *
 * All interface operations that might possibly attempt delivery of data to
 * its destination will fail immediately without further observable action
 * if "error" is already nonzero.  In that case they will return "error".
 *
 * Destruction must attempt to flush any remaining buffered data if "error" is
 * 0 at that time.  However, it is not clear who if anyone will be notified of
 * any failure during such an attempt.  Therefore a careful client may wish to
 * call "flush" before relinquishing its access to this interface.
 *
 * Buffer Management:
 *
 * All buffer allocation is handled by the implementation.
 *
 * The client may write data to [current, end) and advance "current" as far
 * as "end".  Doing so appends to the conceptual stream and therefore makes
 * the implementation responsible for the portion of the buffer over which
 * the client advanced the "current" pointer.
 *
 * The implementation may make any change to the buffer that keeps
 * current <= end and keeps [current, end) writable, so long as it retains
 * responsibility for data already appended to the conceptual stream.
 *
 * Interface Lifetime: single-owner.
 * Thread Safety: requires synchronization.
 */
typedef struct vrpsx_ostreambuf_struct {
#define VRPSX_OSTREAMBUF_MEMBERS(VTBL)                          \
    VTBL const *vtbl;                                           \
                                                                \
    /* The current write position; see "Buffer Management".     \
     */                                                         \
    char *current;                                              \
                                                                \
    /* The end of the writable region; see "Buffer Management". \
     *                                                          \
     * The interface client must not modify this member.        \
     */                                                         \
    char *end;                                                  \
                                                                \
    /* The errno-like error number associated with the first    \
     * failure to deliver data to its destination, or 0 if as   \
     * yet there have been no such failures (possibly because   \
     * as yet there have been no delivery attempts).            \
     *                                                          \
     * The interface client must not modify this member.        \
     */                                                         \
    int error;

    VRPSX_OSTREAMBUF_MEMBERS(struct vrpsx_ostreambuf_vtbl_struct)
} vrpsx_ostreambuf_t;

typedef struct vrpsx_ostreambuf_end_struct {
} vrpsx_ostreambuf_end_t;

typedef struct vrpsx_ostreambuf_vtbl_struct {
#define VRPSX_OSTREAMBUF_VTBL_MEMBERS(TYPE)                             \
                                                                        \
    /* Flushes, destroys, and deallocates this interface (and the       \
     * underlying object and its buffer space, as appropriate).         \
     *                                                                  \
     * As appropriate, destruction occurs even if flushing fails,       \
     * and deallocation occurs even if destruction fails,               \
     */                                                                 \
    int (*disown)(                                                      \
        TYPE **osb);                                                    \
                                                                        \
    /* Attempts to send all buffered data to its destination,           \
     * though that destination may have its own buffers that            \
     * are NOT flushed by this operation.                               \
     *                                                                  \
     * The implementation may modify "current" and/or "end".            \
     *                                                                  \
     * Fails immediately of "error" is nonzero.  Otherwise sets "error" \
     * on failure to deliver data to its destination.  Returns "error"  \
     * in the case of such failure.                                     \
     *                                                                  \
     * No other errors are required.                                    \
     *                                                                  \
     * Returns the new value of the "error" member.                     \
     */                                                                 \
    int (*flush)(                                                       \
        TYPE *osb);                                                     \
                                                                        \
    /* Equivalent To:                                                   \
     *      osb->vtbl->putn(osb, &ch, 1)                                \
     */                                                                 \
    int (*put1)(                                                        \
        TYPE *osb,                                                      \
        char ch);                                                       \
                                                                        \
    /* Appends [begin, begin + size) to the conceptual stream.  That memory \
     * region must NOT overlap the buffer presented by this interface.  \
     *                                                                  \
     * The implementation might flush any initial region of the         \
     * buffered data, and it may modify "current" and/or "end".         \
     *                                                                  \
     * Fails immediately if "error" is nonzero.  Otherwise sets "error" \
     * on failure to deliver data to its ultimate destination.  Returns \
     * "error" in the case of such failure.                             \
     *                                                                  \
     * No other errors are required, though an implementation might for \
     * example choose to return EFAULT if (begin == NULL && size > 0).  \
     * But that is not required.                                        \
     *                                                                  \
     * Returns the new value of the "error" member.                     \
     */                                                                 \
    int (*putn)(                                                        \
        TYPE *osb,                                                      \
        char const *begin,                                              \
        size_t size);                                                   \
                                                                        \
    /* Marks the end of this interface, after which come any extension  \
     * operations.  This is not just for documentation: this member     \
     * may trigger a compiler warning or error if you make a mistake    \
     * while listing methods, though you may need to enable pointer     \
     * incompatibility checks in your compiler.                         \
     */                                                                 \
    vrpsx_ostreambuf_end_t const *ostreambuf_end;

    VRPSX_OSTREAMBUF_VTBL_MEMBERS(vrpsx_ostreambuf_t)
} vrpsx_ostreambuf_vtbl_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_OSTREAMBUF_H */
